--
-- Speciaziation class for belt rake
-- V1.0
-- @author 50keda
-- @date 20/1/2017
-- Copyright (C) 50keda, All Rights Reserved.


BeltRake = {}
BeltRake.States = {STARTING = 0, RUNNING = 1, STOPPING = 2, STOPPED = 3}
BeltRake.Modes = {TEDDER = 0, WINDROWER = 1}

function BeltRake.prerequisitesPresent(specializations)
	return 	SpecializationUtil.hasSpecialization(AnimatedVehicle, specializations) and 
			SpecializationUtil.hasSpecialization(Attachable, specializations) and
			SpecializationUtil.hasSpecialization(WorkArea, specializations)
end;

function BeltRake:load(savegame)
	
	self.getIsWorkAreaActiveOld = WorkArea.getIsWorkAreaActive
	self.getIsWorkAreaActive = BeltRake.getIsWorkAreaActive
	self.isDetachAllowed = Utils.overwrittenFunction(self.isDetachAllowed, BeltRake.isDetachAllowed);
	
	self.switchToMode = BeltRake.switchToMode
	self.updateAnimationSpeed = BeltRake.updateAnimationSpeed
	
	self.beltRakeState = BeltRake.States.STOPPED
	
	local key = "vehicle.beltRake"
	
	-- load initial mode
	if Utils.getNoNil(getXMLString(self.xmlFile, key .. "#initialMode"), "tedder") == "tedder" then
		self.beltRakeMode = BeltRake.Modes.TEDDER
	else
		self.beltRakeMode = BeltRake.Modes.WINDROWER
	end

	-- load fade in out factor
	self.beltRakeFadeIn = Utils.getNoNil(getXMLInt(self.xmlFile, key .. "#fadeIn"), 10)
	self.beltRakeFadeOut = Utils.getNoNil(getXMLInt(self.xmlFile, key .. "#fadeOut"), 25)

	-- load switch animations
	self.beltRakeTSAnimName = getXMLString(self.xmlFile, key .. ".tedderSwitch#animationName")
	self.beltRakeWSAnimName = getXMLString(self.xmlFile, key .. ".windrowerSwitch#animationName")
	
	-- load clip animations
	self.beltRakeClipAnims = {}
	local clipAnimI = 0
	while true do
		
		local clipAnimKey = key .. string.format(".clipAnimations.animation(%d)", clipAnimI)

		if not hasXMLProperty(self.xmlFile, clipAnimKey) then
			break
		end

		local clipAnim = {}
		local initialized = false
		clipAnim.node = Utils.indexToObject(self.components, getXMLString(self.xmlFile, clipAnimKey .. "#rootNode"))

		if clipAnim.node ~= nil then
			
			clipAnim.charSet = getAnimCharacterSet(clipAnim.node)
			
			if clipAnim.charSet ~= nil then
				local clipName = getXMLString(self.xmlFile, clipAnimKey .. "#clipName")
				local clipIndex = getAnimClipIndex(clipAnim.charSet, clipName)

				if clipIndex ~= -1 then
					assignAnimTrackClip(clipAnim.charSet, 0, clipIndex)
					setAnimTrackSpeedScale(clipAnim.charSet, 0, 1)
					setAnimTrackLoopState(clipAnim.charSet, 0, true)
					setAnimTrackTime(clipAnim.charSet, 0, 0)

					initialized = true
				else
					print("BeltRake: Clip animation initialization failed, given animation clip: " .. clipName .. " doesn't exists!")
				end
			else
				print("BeltRake: Clip animation initialization failed, given \"rootNode\" doesn't have charSet!")
			end

			clipAnim.maxSpeed = Utils.getNoNil(getXMLFloat(self.xmlFile, clipAnimKey .. "#maxSpeed"), 1.75)
			clipAnim.minSpeed = Utils.getNoNil(getXMLFloat(self.xmlFile, clipAnimKey .. "#minSpeed"), 0.1)
			clipAnim.lastSpeed = 0
		else
			print("BeltRake: Clip animation initialization failed, given \"rootNode\" doesn't exists!")
		end

		if initialized then
			table.insert(self.beltRakeClipAnims, clipAnim)
		end

		clipAnimI = clipAnimI + 1
	end

	-- load animations
	self.beltRakeAnims = {}
	local animI = 0
	while true do
		
		local animKey = key .. string.format(".animations.animation(%d)", animI)

		if not hasXMLProperty(self.xmlFile, animKey) then
			break
		end

		local anim = {}
		local initialized = false
		anim.name = getXMLString(self.xmlFile, animKey .. "#name")

		if anim.name ~= nil then
			
			if self.animations[anim.name] ~= nil then
				anim.maxSpeed = Utils.getNoNil(getXMLFloat(self.xmlFile, animKey .. "#maxSpeed"), 1.75)
				anim.minSpeed = Utils.getNoNil(getXMLFloat(self.xmlFile, animKey .. "#minSpeed"), 0.1)
				anim.lastSpeed = 0
			
				initialized = true
			else
				print(string.format("BeltRake: Animation initialization failed, given animation named: \"%s\" doesn't exists!", anim.name))
			end
		else
			print("BeltRake: Animation initialization failed, animation definition is missing \"name\" attribute!")
		end

		if initialized then
			table.insert(self.beltRakeAnims, anim)
		end

		animI = animI + 1
	end

	if savegame ~= nil then
		self.beltRakeMode = Utils.getNoNil(getXMLInt(savegame.xmlFile, savegame.key.."#beltRakeMode"), self.beltRakeMode)
    end
	
	-- set proper mode
	self:switchToMode(self.beltRakeMode)

	self:addConflictCheckedInput(InputBinding.BELTRAKE_SWITCH_MODE)
end

function BeltRake:getSaveAttributesAndNodes(nodeIdent)
	local attributes = 'beltRakeMode="' .. tostring(self.beltRakeMode) .. '"'
	return attributes, nil
end

function BeltRake:delete()
end

function BeltRake:mouseEvent(posX, posY, isDown, isUp, button)
end

function BeltRake:keyEvent(unicode, sym, modifier, isDown)
end

function BeltRake:isDetachAllowed(superFunc)

	-- if not lowered forbid detaching
	if not self:isLowered() then
		return false
	end

	-- if implement is marked as lowered already but joint still isn't in bottom most position also forbid detaching
 	local implement = self.attacherVehicle:getImplementByObject(self)
    local jointDesc = self.attacherVehicle.attacherJoints[implement.jointDescIndex]
    if jointDesc.allowsLowering and jointDesc.moveDown and jointDesc.moveAlpha ~= jointDesc.lowerAlpha then
        return false
    end

    -- finally call super function for the answer
	if superFunc ~= nil then
		return superFunc(self)
	end

	return true
end

function BeltRake:getIsWorkAreaActive(area)
	
	local isActive = self:getIsWorkAreaActiveOld(area)
	local areaTypeName =  WorkArea.areaTypeIntToName[area.type]

	if not self:isLowered() then
		return false
	end

	if self.beltRakeMode == BeltRake.Modes.TEDDER and (areaTypeName == "tedder" or areaTypeName == "tedderDrop") then
		return isActive
	elseif self.beltRakeMode == BeltRake.Modes.WINDROWER and (areaTypeName == "windrower" or areaTypeName == "windrowerDrop") then
		return isActive
	end

	return false
end

function BeltRake:onTurnedOn(noEventSend)
	self.beltRakeState = BeltRake.States.STARTING
end

function BeltRake:onTurnedOff(noEventSend)
	self.beltRakeState = BeltRake.States.STOPPING
end

function BeltRake:update(dt)

	if self.isClient then

		if self:getIsActive() and self:getIsActiveForInput() then
			if InputBinding.hasEvent(InputBinding.BELTRAKE_SWITCH_MODE) then
				
				local newMode = nil
				if self.beltRakeMode == BeltRake.Modes.TEDDER then
					newMode = BeltRake.Modes.WINDROWER
				else
					newMode = BeltRake.Modes.TEDDER
				end

				self:switchToMode(newMode)

			end
		end

		-- if no animations initialized just skip STOPPING, STARTING and any animation related calculations
		local animsCount = table.getn(self.beltRakeClipAnims) + table.getn(self.beltRakeAnims)
		if animsCount <= 0 then
			
			if self.beltRakeState == BeltRake.States.STOPPING then
				self.beltRakeState = BeltRake.States.STOPPED
			elseif self.beltRakeState == BeltRake.States.STARTING then
				self.beltRakeState = BeltRake.States.RUNNING
			end

		-- otherwise if not stopped make sure to set proper animation speeds and enable/disable them
		elseif self.beltRakeState ~= BeltRake.States.STOPPED then

			-- upon current attacher vehicle motor rpm calculate current rpm percentage used for determinating new speed of animations
			local rpmPercentage = -1
			if self.attacherVehicle ~= nil and self.attacherVehicle.motor ~= nil then
				rpmPercentage = self.attacherVehicle.motor:getLastMotorRpm() / self.attacherVehicle.motor:getMaxRpm()
				rpmPercentage = math.min(1.0, rpmPercentage)
			end

			local stoppedAnimsCount = 0
			local runningAnimsCount = 0

			local currAnimState = nil
			for i, clipAnim in pairs(self.beltRakeClipAnims) do		
				currAnimState = self:updateAnimationSpeed(clipAnim, rpmPercentage, true)
				
				if currAnimState == BeltRake.States.STOPPED then
					stoppedAnimsCount = stoppedAnimsCount + 1
				elseif currAnimState == BeltRake.States.RUNNING then
					runningAnimsCount = runningAnimsCount + 1
				end
			end

			for i, anim in pairs(self.beltRakeAnims) do
				currAnimState = self:updateAnimationSpeed(anim, rpmPercentage, false)
				
				if currAnimState == BeltRake.States.STOPPED then
					stoppedAnimsCount = stoppedAnimsCount + 1
				elseif currAnimState == BeltRake.States.RUNNING then
					runningAnimsCount = runningAnimsCount + 1
				end
			end

			if stoppedAnimsCount == animsCount then
				self.beltRakeState = BeltRake.States.STOPPED
			elseif runningAnimsCount == animsCount then
				self.beltRakeState = BeltRake.States.RUNNING
			end
		end

	end
end

function BeltRake:updateAnimationSpeed(animTable, rpmPercentage, isClipAnimation)
	local currentSpeed = animTable.lastSpeed
	local newSpeed = animTable.maxSpeed -- set desired speed to max speed
	local newState = self.beltRakeState
	
	-- if rpms exist then do set proper new speed
	if rpmPercentage ~= -1 then
		newSpeed = animTable.maxSpeed * rpmPercentage
		newSpeed = math.max(newSpeed, animTable.minSpeed)
	end

	-- depending on current state ease in/out to running speed or to stopped state
	if self.beltRakeState == BeltRake.States.STARTING then
		if currentSpeed < newSpeed then
			newSpeed = currentSpeed + animTable.minSpeed / self.beltRakeFadeIn
		else
			newState = BeltRake.States.RUNNING
		end
	elseif self.beltRakeState == BeltRake.States.STOPPING then
		if currentSpeed > animTable.minSpeed / 100 then
			local divider = self.beltRakeFadeOut
			-- stop it quicker as it's in work position aka lowered and unfolded
			if self:isLowered() and not self:getIsUnfolded() then
				divider = divider / 2
			end
			newSpeed = currentSpeed - animTable.minSpeed / divider
		else
			newSpeed = 0
			newState = BeltRake.States.STOPPED						
		end
	end
	
	if newSpeed > 0 then
		if newSpeed ~= currentSpeed then
			if isClipAnimation then
				setAnimTrackSpeedScale(animTable.charSet, 0, newSpeed)
				enableAnimTrack(animTable.charSet, 0)
			else
				self:playAnimation(animTable.name, newSpeed, self:getAnimationTime(animTable.name))
			end
		end
	else
		if isClipAnimation then
			disableAnimTrack(animTable.charSet, 0)
		else
			self:stopAnimation(animTable.name)
		end
	end

	animTable.lastSpeed = newSpeed
	return newState
end

function BeltRake:draw()
	if self.isClient and self:getIsActiveForInput(true) then
		if self.beltRakeMode == BeltRake.Modes.TEDDER then
    		g_currentMission:addHelpButtonText(g_i18n:getText("action_BELTRAKE_TO_WINDROWER"), InputBinding.BELTRAKE_SWITCH_MODE, nil, GS_PRIO_HIGH)
		else
			g_currentMission:addHelpButtonText(g_i18n:getText("action_BELTRAKE_TO_TEDDER"), InputBinding.BELTRAKE_SWITCH_MODE, nil, GS_PRIO_HIGH)
		end
	end
end

function BeltRake:switchToMode(mode, noEventSend)
	if mode == BeltRake.Modes.TEDDER then
		if self.beltRakeTSAnimName ~= nil then
			self:playAnimation(self.beltRakeTSAnimName, 1, self:getAnimationTime(self.beltRakeTSAnimName))
		end
	else
		if self.beltRakeWSAnimName ~= nil then
			self:playAnimation(self.beltRakeWSAnimName, 1, self:getAnimationTime(self.beltRakeWSAnimName))
		end
	end

	self.beltRakeMode = mode

	BeltRakeModeEvent.sendEvent(self, mode, noEventSend)
end

--
-- MP ready
--

BeltRakeModeEvent = {}
BeltRakeModeEvent_mt = Class(BeltRakeModeEvent, Event)

InitEventClass(BeltRakeModeEvent, "BeltRakeModeEvent")

function BeltRakeModeEvent:emptyNew()
    local self = Event:new(BeltRakeModeEvent_mt)
    return self
end

function BeltRakeModeEvent:new(object, newBeltRakeMode)
    local self = BeltRakeModeEvent:emptyNew()
    self.object = object
	self.newBeltRakeMode = newBeltRakeMode
    return self
end

function BeltRakeModeEvent:readStream(streamId, connection)
    self.object = readNetworkNodeObject(streamId);
	self.newBeltRakeMode = streamReadInt8(streamId)
	self:run(connection)
end

function BeltRakeModeEvent:writeStream(streamId, connection)
	writeNetworkNodeObject(streamId, self.object)
	streamWriteInt8(streamId, self.newBeltRakeMode)
end

function BeltRakeModeEvent:run(connection)
	if not connection:getIsServer() then
        g_server:broadcastEvent(BeltRakeModeEvent:new(self.object, self.newBeltRakeMode), nil, connection, self.object)
    end
    if self.object ~= nil then
    	self.object:switchToMode(self.newBeltRakeMode, true)
	end
end

function BeltRakeModeEvent.sendEvent(vehicle, newBeltRakeMode, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			g_server:broadcastEvent(BeltRakeModeEvent:new(vehicle, newBeltRakeMode), nil, nil, vehicle)
		else
			g_client:getServerConnection():sendEvent(BeltRakeModeEvent:new(vehicle, newBeltRakeMode))
		end
	end
end
